#!/usr/bin/env python
#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
import sys
import os
from time import gmtime, strftime
import subprocess
import shutil
from pprint import pprint

from _fpga import _fpga

from _mysigs import Clock,Reset

_xilinx_default_pin_attr = {
    "NET": None,
    "LOC" : None,
    "IOSTANDARD": None,
    "SLEW": None,
    "DRIVE": None
    }
 
class _ise(object):
    """ """
    
    def __init__(self, top_name=None, path=None):
        self._path = path if path is not None else '.'
            
        self._fpga = None
        self.tcl_name = None
        self.top_name = None
        self.hdl_files = None
        self.hdl_file_list = []
        
        if top_name:
            if isinstance(top_name, str):
                self.top_name = top_name
                self.tcl_name = top_name + '.tcl'
            else:
                raise TypeError('top_name needs to be string')
 
        # start with the text string for the TCL script
        self.tcl_script = '#\n#\n# ISE implementation script\n'
        date_time = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
        self.tcl_script += '# create: %s\n' % \
                           date_time
        self.tcl_script += '# by: %s\n' % \
                           os.path.basename(sys.argv[0])
        self.tcl_script += '#\n#\n'

    def add_files(self, fn):
        if isinstance(fn, (list,tuple)):
            for ff in fn:
                if not isinstance(ff,str):
                    raise TYpeError("list or tuple entry needs to be a string")
                if ff not in self.hdl_file_list:
                    self.hdl_file_list.append(ff)
        elif isinstance(files,str):
            if ff not in self.hdl_file_list:
                self.hdl_file_list.append(files)
        
    def create_tcl(self, filename=None):
        """Create the project TCL script
        """
        print(self._path,filename,self.tcl_name)
        if filename:
            fn = os.path.join(self._path, filename)
        else:
            fn = os.path.join(self._path, self.tcl_name)
 
        self.tcl_script += '# set compile directory:\n'
        #self.tcl_script += 'set compile_directory %s\n'%self._path
        self.tcl_script += 'set compile_directory %s\n'%'.'

        if self.top_name:
            self.tcl_script += 'set top_name %s\n'%self.top_name
            self.tcl_script += 'set top %s\n'%self.top_name

        # @todo: add one file at a time, explicit
        #-#if self.hdl_files:
        #-#    self.tcl_script += '# input source files:\n'
        #-#    self.tcl_script += 'set hdl_files [ list \\\n'
        #-#    for f in self.hdl_file_list:
        #-#        self.tcl_script += ' '*17
        #-#        self.tcl_script += '%s \\\n'%f
        #-#        self.tcl_script += ']\n'
 
 
        self.tcl_script += '# set Project:\n'
        self.tcl_script += 'set proj %s\n'% self.top_name
 
        # @note: because the directory is changed everything
        #        is relative to self._path
        self.tcl_script += '# change to the directory:\n'
        self.tcl_script += 'cd %s\n'% self._path
 
        # @todo: verify UCF file exists
        bdir,ucffn = os.path.split(self.fpga.ucf_file)
        self.tcl_script += '# set ucf file:\n'
        self.tcl_script += 'set constraints_file %s\n'% (ucffn)

        #-## test whether ise project file exits
        #-## @todo: check the Xilinx version, if greater than ??? use xise
        #-##        else use .ise
        #-#f = os.path.join(self._path, self.top_name + '.xise')
        #-#if os.path.exists(f):
        #-#    os.remove(f)

        self.tcl_script += '# set variables:\n'
        pj_fn = self.top_name + '.xise'
        # Create or open an ISE project (xise?)
        print('Project name : %s ' % (pj_fn))
        pjfull = os.path.join(self._path,pj_fn)

        # let the TCL file be the master file, always create
        # a new project.  If a user uses this to "bootstrap"
        # the need to take care to rename the project if modfied
        # else it will be overwritten.
        if os.path.isfile(pjfull):
            os.remove(pjfull)
        #    self.tcl_script += 'project open %s \n' % (pj_fn)
        #else:
        self.tcl_script += 'project new %s\n' % (pj_fn)
    
        if self.fpga.family:
            self.tcl_script += 'project set family %s\n'%self.fpga.family
            self.tcl_script += 'project set device %s\n'%self.fpga.device
            self.tcl_script += 'project set package %s\n'%self.fpga.package
            self.tcl_script += 'project set speed %s\n'%self.fpga.speed
 
        # add the hdl files
        self.tcl_script += '\n'
        self.tcl_script += '# add hdl files:\n'
        self.tcl_script += 'xfile add %s\n'%(ucffn)
        for hdl_file in self.hdl_file_list:
            self.tcl_script += 'xfile add %s\n'%hdl_file
 
        self.tcl_script += '# test if set_source_directory is set:\n'
        self.tcl_script += 'if { ! [catch {set source_directory'
        self.tcl_script += ' $source_directory}]} {\n'
        self.tcl_script += '  project set "Macro Search Path"\n'
        self.tcl_script += ' $source_directory -process Translate\n'
        self.tcl_script += '}\n'
 
        # @todo : need an elgent way to manage all the insane options, 90% 
        #         of the time the defaults are ok, need a config file or 
        #         something to overwrite.  These should be in a dict or
        #         refactored or something
        self.tcl_script += "project set \"FPGA Start-Up Clock\" \"JTAG Clock\" -process \"Generate Programming File\" \n"

        # run the implementation
        self.tcl_script += '# run the implementation:\n'
        self.tcl_script += 'process run "Synthesize" \n'
        self.tcl_script += 'process run "Translate" \n'
        self.tcl_script += 'process run "Map" \n'
        self.tcl_script += 'process run "Place & Route" \n'
        self.tcl_script += 'process run "Generate Programming File" \n'
        # close the project
        self.tcl_script += '# close the project:\n'
        self.tcl_script += 'project close\n' 
 
        fid = open(fn, 'w')
        fid.write(self.tcl_script)
        fid.close() 

    def run(self, filename=None):
        """Run the created TCL script """
        if filename:
            tcl_name = filename
        else:
            tcl_name = os.path.join(self._path, self.tcl_name)
 
        if not os.path.exists(self._path):
            os.mkdir(self._path)
 
        cmd = ['xtclsh', tcl_name]
        print('running command: %s'%(cmd))
        try:
            txt = subprocess.check_call(cmd,
                                        stderr=subprocess.STDOUT)
        except Exception, err:
            print(err)
            raise err

    @property
    def fpga(self):
        return self._fpga
    @fpga.setter
    def fpga(self, fpga):      
        self._fpga = fpga

  
class XilinxFPGA(_fpga):
    def __init__(self, prj_name='', path=None):
        self._path = path if path is not None else '.'
        
        self.ucf_file = None
        self.family = None
        self.device = None
        self.package = None
        self.speed = None
        _fpga.__init__(self)
        self.top_name = prj_name

        # @todo: verify this is the correct intended usage, i.e. 
        #        the file list doesn't already exist somewhere
        self.hdl_file_list = []
                                      
    @property
    def path(self):
        return self._path
    @path.setter
    def path(self,p):
        self._path = p
        
    def set_device(self, family, device, package, speed):
        
        if not isinstance(family, str):
            raise TypeError('"Family" needs to be a string')
        if not isinstance(device, str):
            raise TypeError('"Device" needs to be a string')
        if not isinstance(speed, str):
            raise TypeError('"Speed" needs to be a string')
        if not isinstance(package, str):
            raise TypeError('"Package" needs to be a string')
      
        self.family = family
        self.device = device
        self.package = package
        self.speed = speed
            
      
    def create_ucf(self, filename='my.ucf'):
        self.ucf_file = os.path.join(self._path, filename)
        ustr = ""
        #ustr = '# UCF file automatically create by "%s"\n'% \
        #       os.path.basename(sys.argv[0])

        print(type(self.ports),self.ports)
        ustr += "#\n"
        for port_name,port in self.ports.items():
            if port.inuse:
                _pins = port.pins

                for ii,pn in enumerate(_pins):
                    if len(_pins) == 1:
                        ustr += "NET \"%s\" " % (port_name)
                    else:
                        ustr += "NET \"%s<%d>\" " % (port_name,ii)

                    # pure numeric pins need a preceeding "p" otherwise
                    # use the string defined
                    if isinstance(pn,str):
                        ustr += "LOC = \"%s\" " % (str(pn))
                    else:
                        ustr += "LOC = \"p%s\" " % (str(pn))

                    # additional pin parameters
                    for kp,vp in port.kwargs.items():
                        if kp.lower() in ("pullup",) and vp is True:
                            ustr += " | %s " % (kp)
                        else:
                            ustr += " | %s = %s " % (kp,vp)
                    ustr += ";\n"
                
        ustr += "#\n"

        # @todo: loop through the pins again looking for clocks
        for port_name,port in self.ports.items():
            if port.inuse and isinstance(port.sig,Clock):
                period = 1/(port.sig.frequency/1e9)
                ustr += "NET \"%s\" TNM_NET = \"%s\"; \n" % (port_name,port_name)
                ustr += "TIMESPEC \"TS_%s\" = PERIOD \"%s\" %.7f ns HIGH 50%%;" \
                    % (port_name,port_name,period)
                ustr += "\n"
        ustr += "#\n"

        fid = open(self.ucf_file, 'w')
        fid.write(ustr)
        fid.close()      
        print(ustr)

    def run(self, use='verilog'):
        """Create the files and run the FPGA toolflow
        This function will convert the MyHDL design, create the 
        constraint files, and run the FPGA toolflow.
        """
        print(self.top_name)
        use = use.lower()
        ext = {'verilog':'.v', 'vhdl':'vhd'}
        
        # convert the design to the intermidiate V* 
        self._scrub(self.hdl_file_list)
        cfiles = self.convert(to=use)
        for ff in cfiles:            
            self.hdl_file_list.append(ff)
        
        # create the constraints and project files
        self.create_ucf(filename=self.name+'.ucf')
        self._tool_chain = _ise(top_name=self.top_name,path=self._path)
        self._tool_chain.fpga = self
        pprint(vars(self._tool_chain.fpga))
        
        # add the files to the toolchain
        self._tool_chain.add_files(self.hdl_file_list)

        # create the TCL file and run the tools
        self._tool_chain.create_tcl()
        #self._tool_chain.run()

    def rerun(self):
        """Run the FPGA toolflow using the existing files
        """
        self._tool_chain = _ise(top_name=self.top_name,path=self._path)
        self._tool_chain.fpga = self
        pprint(vars(self._tool_chain.fpga))
        self._tool_chain.run()
        
    def __repr__(self):
        s = 'FPGA:\n'
        if self.family:
          s += 'Family: %s\n'% self.family
        s += 'Device: %s\n'% self.device
        s += 'Package: %s\n'% self.package
        s += 'Speed: %s\n'% self.speed
        return s
 



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import shutil
import myhdl as _myhdl

class Simulator:
    """
    This class is intended to be used with "testbench" method
    of testing converted MyHDL modules.  The "testbench" method
    refers to a convertible testbench.  The same checks that
    are performed in the python code is performed in V* code.
    """
    
    vhdl_dir = 'vhdl/'
    verilog_dir = 'verilog/'
    build_dir = 'build/'
    
    def __init__(self):
        self.testbench = None
        self.hdl_path = None

    def _get_pck(self):
        ver = _myhdl.__version__
        ver = ver.replace('.', '')[:2]
        pck = "pck_myhdl_%s.vhd" % (ver)
        return pck
    
    def convert(self, testbench):
        self.testbench = testbench

    def compile(self):
        pass

    def run(self):
        pass

    def all(self, testbench):
        self.convert(testbench)
        self.compile()
        self.run()
        

# need a better name??
class Myhdl(Simulator):
    name = 'myhdl'
    def run(self):
        assert self.testbench is not None
        _myhdl.Simulation(self.testbench()).run()
        
class Ghdl(Simulator):
    name = 'ghdl'
    def convert(self, testbench):
        assert testbench is not None
        self.testbench = testbench
        tb = testbench
        
        self.hdl_path = os.path.join(self.vhdl_dir, tb.func_name+'.vhd')
        pck = self._get_pck()
        self.pck_path = os.path.join(self.vhdl_dir, pck)
        if not os.path.isdir(self.vhdl_dir):
            os.mkdir(self.vhdl_dir)
        if os.path.isfile(self.hdl_path):
            os.remove(self.hdl_path)
        _myhdl.toVHDL(tb)
        shutil.move(tb.func_name+'.vhd', self.vhdl_dir)
        if os.path.isfile(self.pck_path):
            os.remove(self.pck_path)
        shutil.move(pck, self.pck_path)
        
    def compile(self):
        assert self.testbench is not None
        pck = self._get_pck()
        subprocess.check_output(['ghdl', '-a', '%s' % (self.pck_path)],
                                stderr=subprocess.STDOUT)
        subprocess.check_output(['ghdl', '-a', '%s' % (self.hdl_path)],
                                stderr=subprocess.STDOUT)
        subprocess.check_output(['ghdl', '-e', '%s' % (self.testbench.func_name)],
                                stderr=subprocess.STDOUT)

    def run(self):
        try:
            stext = subprocess.check_output(['ghdl', '-r',
                                             '%s' % (self.testbench.func_name)],
                                            stderr=subprocess.STDOUT)
        except Exception, err:
            stext = err.output
            if 'End of Simulation' in stext:
                pass # continue all is ok
            else:
                raise err
            
        if 'AssertionError' in stext:
            raise StandardError('VHDL simulation failed')

class Icarus(Simulator):        
    name='icarus'
    def convert(self, testbench):
        assert testbench is not None
        self.testbench = testbench
        tb = testbench
        
        self.hdl_path = os.path.join(self.verilog_dir, tb.func_name+'.v')
        if not os.path.isdir(self.verilog_dir):
            os.mkdir(self.verilog_dir)
        if os.path.isfile(self.hdl_path):
            os.remove(self.hdl_path)
        _myhdl.toVerilog(tb)
        shutil.move(tb.func_name+'.v', self.verilog_dir)

    def compile(self):
        self.out = self.hdl_path[:-2]
        subprocess.check_output(['iverilog',
                                 '-o',
                                 '%s' % (self.out),
                                 '%s' % (self.hdl_path)],
                                stderr=subprocess.STDOUT)
        
    def run(self):
        stext = subprocess.check_output([self.out],
                                        stderr=subprocess.STDOUT)
        if 'AssertionError' in stext:
            raise StandardError('Verilog simulation failed')


AllSims = [Myhdl(), Ghdl(), Icarus()]

def run_all(test):
    for sim in AllSims:
        try:
            test(sim)
            print('   %-8s:%s: success' % (sim.name, test.func_name))
        except Exception, err:
           print('   %-8s:%s: failed %s' % (sim.name, test.func_name, err))





from myhdl import *
from myhdl_tools import Clock,Reset

def test_reset():

    def _test():
        clock = Clock(0)
        reset = Reset(1, active=0, async=True)
        
        Pulses = [10, (10,), (10,10), (10,10,10)]
        pcnt = Signal(0)
        rtst = Signal(-1)
        
        tbclk = clock.gen()
        
        @instance
        def tbmon():
            while True:
                yield reset.negedge
                pcnt.next = pcnt + 1
            
        @instance
        def tbstim():
            yield delay(10)
            ts = now()
            for ii,pp in enumerate(Pulses):
                rtst.next = rtst + 1
                yield reset.pulse(pp)
                yield delay(10)
            te = now()
            assert te-ts == 70+10*len(Pulses)
            yield delay(10)
            assert pcnt == 4, "%d"%(pcnt)

            raise StopSimulation

        return tbclk, tbmon, tbstim

    Simulation(_test()).run()

        
if __name__ == '__main__':
    test_reset()



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from myhdl import Signal,intbv
from ... import Clock,Reset
from ..._xilinx import XilinxFPGA

atlys = XilinxFPGA('atlys')
atlys.add_port('clock', Clock(0, frequency=50e6), pins=('L15',))

# on-board LEDs
atlys.add_port('led', Signal(intbv(0)[8:]),
               pins=('U18', 'M14', 'N14', 'L14',
                     'M13', 'D4',  'P16', 'N12'))
# on-board push buttons
atlys.add_port('btn', Signal(intbv(0)[8:]),
               pins=('T15', 'N4', 'P4', 'P3',
                     'F6', 'F5',))
# device
atlys.set_device('spartan6', 'XC6SLX45', 'CSG324', '-3')




from _xboard_list import xbrd
#from _aboard_list import abrd

# list of boards with default pin mappings

def get_boards(name=None):
    if xbrd.has_key(name):
        pass
    else:
        raise ValueError("Invalid board %s"%(name))



#
# Copyright (c) 2013 Alexander Hungenberg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from myhdl import Signal, intbv
from ... import Clock
from ..._xilinx import XilinxFPGA

VCCO = 'LVCMOS33'

pone = XilinxFPGA('pone')
pone.set_device('spartan3E', 'xc3s500e', 'VQ100', '-4')
pone.add_port('clock', Clock(0, frequency=32e6), 89, iostandard='LVCMOS25')
pone.add_port('rx', Signal(False), 88, iostandard='LVCMOS33')
pone.add_port('tx', Signal(False), 90, iostandard='LVCMOS33', drive='4', slew='SLOW')

pone.add_port('wingA', Signal(intbv(0)[16:]),
              (18, 23, 26, 33, 35, 40, 53, 57, 60, 62,
               65, 67, 70, 79, 84, 86),
              iostandard=VCCO)
pone.add_port('wingB', Signal(intbv(0)[16:]),
              (85, 83, 78, 71, 68, 66, 63, 61, 58, 54,
               41, 36, 34, 32, 25, 22),
              iostandard=VCCO)
pone.add_port('wingC', Signal(intbv(0)[16:]),
              (91, 92, 94, 95, 98, 2, 3, 4, 5, 9,
               10, 11, 12, 15, 16, 17),
              iostandard=VCCO)

pone.add_port('jtag_tms', Signal(False), 75, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('jtag_tck', Signal(False), 77, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('jtag_tdi', Signal(False), 100, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('jtag_tdo', Signal(False), 76, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('flash_cs', Signal(False), 24, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('flash_ck', Signal(False), 50, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('flash_si', Signal(False), 27, iostandard='LVTTL', drive='8', slew='fast')
pone.add_port('flash_so', Signal(False), 44, iostandard='LVTTL', drive='8', slew='fast', pullup=True)


#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from myhdl import Signal,intbv
from ... import Clock,Reset
from ..._xilinx import XilinxFPGA

# this file contains the definitions for the Xess Xula
# FPGA development boards: http://www.xess.com

# Xula board by Xess, http://www.xess.com/prods/prod048.php
xula = XilinxFPGA('xula')
xula.add_port('clock', Clock(0, frequency=12e6), pins=(43,))
xula.add_port('chan', Signal(intbv(0)[32:]),
             pins=(36,37,39,50,52,56,57,61,  # 0-7
                   62,68,72,73,82,83,84,35,  # 8-15
                   34,33,32,21,20,19,13,12,  # 17-23
                   07,04,03,97,94,93,89,88)) # 24-31
xula.add_port('chan_clk', Clock(0, frequency=1e6), pins=(44,))
xula.set_device('spartan3A','xc3s200a','VQ100','-4')


xula2 = XilinxFPGA('xula2')
xula2.set_device('spartan6','xc6s1200','VQ100','-4')
xula2.add_port('clock', Clock(0, frequency=12e6), pins=('A9',))
xula2.add_port('chan', Signal(intbv(0)[32:]),
               pins=('R7','R15','R16','M15','M16','K15',  #0-5
                     'K16','J16','J14','F15','F16','C16', #6-11
                     'C15','B16','B15','T4','R2','R1',    #12-17
                     'M2','M1','K3','J4','H1','H2',       #18-23
                     'F1','F2','E1','E2','C1','B1',       #24-29
                     'B2','A2',) )                        #30-31
xula2.add_port('chan_clk', Clock(0, frequency=1e6), pins=('T7',))




import os
import shutil
from glob import glob

import pytest

from myhdl import *
from myhdl_tools import Clock,Reset
from myhdl_tools.boards import get_xilinx_board

def m_blink(clock,reset,toggle):
    MAX_CNT = int(clock.frequency)
    cnt = Signal(intbv(0,min=0,max=MAX_CNT))
    @always_seq(clock.posedge,reset=reset)
    def hdl():
        if cnt == MAX_CNT-1:
            cnt.next = 0
            toggle.next = not toggle
        else:
            cnt.next = cnt + 1
    return hdl

@pytest.fixture
def clean():
    if os.path.isdir('xilinx/'):
        shutil.rmtree('xilinx/')
    if os.path.isdir('altera/'):
        shutil.rmtree('altera/')
    vlist = glob("*.v*")
    for ff in vlist:
        os.remove(ff)


def test_atlys_vhdl(clean):
    brd = get_xilinx_board('atlys', top=m_blink)
    brd.add_port('toggle', Signal(bool(0)), pins=('U18',)) 
    brd.add_port('reset', Reset(0,active=0,async=True), pins=('T15',))
    brd.run('vhdl')

def test_atlys_vhdl_noclean():
    test_atlys_vhdl(None)

def test_atlys_verilog(clean):
    brd = get_xilinx_board('atlys', top=m_blink)
    brd.add_port('toggle', Signal(bool(0)), pins=('U18',)) 
    brd.add_port('reset', Reset(0,active=0,async=True), pins=('T15',))
    brd.run('verilog')

def test_atlys_verilog_noclean():
    test_atlys_verilog(None)

def test_xula_vhdl(clean):
    brd = get_xilinx_board('xula', top=m_blink)
    brd.add_port('toggle', Signal(bool(0)), pins=(36,)) # not a default pin
    brd.add_port('reset', Reset(0,active=0,async=True), pins=(39,))
    brd.run('vhdl')

def test_xula_verilog(clean):
    brd = get_xilinx_board('xula', top=m_blink)
    brd.add_port('toggle', Signal(bool(0)), pins=(36,)) # not a default pin
    brd.add_port('reset', Reset(0,active=0,async=True), pins=(39,))
    brd.run('verilog')


if __name__ == '__main__':
    tl = (test_xula_vhdl, test_xula_verilog, 
          test_atlys_vhdl, test_atlys_verilog,
          test_xula_vhdl, test_xula_verilog, 
          test_atlys_vhdl, test_atlys_verilog)

    for tt in tl:
        print("*"*68)
        print("*"*68)
        clean()
        tt(None)
        print("*"*68)

    #clean()
    #test_xula_vhdl(None)
    #clean()
    #test_xula_verilog(None)
    #
    #clean()
    #test_atlys_vhdl(None)
    #clean()
    #test_atlys_verilog(None)



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import myhdl
from myhdl import instance, delay

ClockList = []

class Clock(myhdl.SignalType):    
    def __init__(self, val, frequency=1, timescale='1ns'):
        self._frequency = frequency
        self._period = 1/frequency
        self._timescale = timescale
        self._set_hticks()
        myhdl.SignalType.__init__(self, bool(val))
        ClockList.append(self)

    @property
    def timescale(self):
        return self._timescale
    @timescale.setter
    def timescale(self, t):
        self._timescale = t
        
    @property
    def frequency(self):
        return self._frequency
    @frequency.setter
    def frequency(self, f):
        self._frequency = f
        self._period = 1/f
        self._set_hticks()
        
    @property
    def period(self):
        return self._period

    def _set_hticks(self):
        #self._nts = self._convert_timescale(self._timescale)
        #self._hticks = int(round(self._period/self._nts))
        self._hticks = 3

    def _convert_timescale(self, ts):
        # @todo: need to complete this, ts is in the form
        #        "[0-9]*["ms","us","ns","ps"], parse the text
        #        format and retrieve a numerical value
        # separate the numerical and text        
        nts = 1e9
        return nts

    def gen(self, hticks=None):
        if hticks is None:
            hticks = self._hticks
        else:
            self._hticks = hticks

        #print('hticks %d'%(hticks))
        @instance
        def gclock():
            self.next = False
            while True:
                yield delay(hticks)
                self.next = not self.val

        return gclock

class Reset(myhdl.ResetSignal):
    def __init__(self, val, active, async):
        myhdl.ResetSignal.__init__(self,val,active,async)

    def pulse(self, delays=10):
        if isinstance(delays,(int,long)):
            self.next = self.active
            yield delay(delays)
            self.next = not self.active
        elif isinstance(delays,tuple):
            assert len(delays) in (1,2,3), "Incorrect number of delays"
            self.next = not self.active if len(delays)==3 else self.active
            for dd in delays:
                yield delay(dd)
                self.next = not self.val
            self.next = not self.active
        else:
            raise ValueError("%s type not supported"%(type(d)))
        
            






from myhdl import *
from myhdl_tools import Clock

def test_clock():

    def _test():
        clock1 = Clock(0, frequency=50e6)
        clock2 = Clock(0, frequency=100e6)
        clock3 = Clock(1, frequency=12.5e6)

        # the clock module will default to 3 half period
        # ticks (full period 6 sim ticks) need to set the
        # hticks for the different clocks.
        tbclk1 = clock1.gen(hticks=6)
        tbclk2 = clock2.gen(hticks=3)
        tbclk3 = clock3.gen(hticks=24)

        @instance
        def tbstim():
            for cc,ed in zip((clock1,clock2,clock3),(12,6,48)):
                yield cc.posedge
                ts = now()
                yield cc.posedge
                te = now()
                assert te-ts == ed, "%d != %d"%(te-ts,ed)

            raise StopSimulation
            
            
        return tbclk1, tbclk2, tbclk3, tbstim

    Simulation(_test()).run()


if __name__ == '__main__':
    test_clock()



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import Queue
from myhdl import Signal, delay

class SignalQueue(Queue.Queue):
    def __init__(self, maxsize=0):
        self.Empty = Signal(True)
        self.Full = Signal(False)
        self.maxsize = maxsize
        self._max = 0
        Queue.Queue.__init__(self, maxsize=maxsize)

    def put(self, item, block=True, timeout=None):
        assert timeout is None, "Timeout not yet implemented"
        #print("block %s full %s" % (block, self.Full))
        if block and self.Full:
            yield self.Full.negedge
        Queue.Queue.put(self, item, block=False, timeout=None)
        self._max = max(self._max,self.qsize())
        self.Empty.next = False
        if self.qsize() == self.maxsize and self.maxsize > 0:
            self.Full.next = True
        if self.Empty or not self.Full:
            yield delay(1)

    def put_nowait(self, item):
        self.put(item, Block=False, timeout=None)

    def get(self, item, block=True, timeout=None):
        assert timeout is None, "Timeout not yet implemented"
        #print("block %s empty %s %d" % (block, self.Empty, self.qsize()))
        if block and self.Empty:
            yield self.Empty.negedge
        item[0] = Queue.Queue.get(self, block=False, timeout=None)
        if self.qsize() == 0:
            self.Empty.next = True
        self.Full.next = False
        if not self.Empty or self.Full:
            yield delay(1)        
    
    def get_nowait(self, item):
        yield self.get(item, block=False, timeout=None)

    def __str__(self):
        s = 'len %d, max %d'%(self.qsize(),self._max)
        return s

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return self.qsize()



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The following is a collection of different Xilinx based FPGA
development boards.  The defualt pin names (from schematics)
is defined.  If an HDL top-level matches the hardware pin
names no ports need to be defined.

For each pin defined for the hardware a port can be defined,
the tools will map the HDL top-level ports to the ports defined
below or the ports can be overridden with the HDL names, see
examples for port redefinitions
"""
import inspect

from .._xilinx import XilinxFPGA
import xilinx

# Board list
xbrd = {}
for m in inspect.getmembers(xilinx):
    if isinstance(m[1], XilinxFPGA):
        xbrd[m[0]] = m[1]
        
def get_xilinx_board(name=None, top=None, path='./xilinx/'):
    brd = None
    if xbrd.has_key(name):
       brd = xbrd[name]
    else:
        # @todo: look at a name and determine if it is close to one
        #        of the existing boards and provide a suggestion
        print('boards: %s'%(str(xbrd.keys)))
        raise ValueError('Invalid board %s'%(name))

    if top is not None:
        brd.set_top(top)
    brd.path = path

    return brd




# Xess corp boards
from _xula import xula,xula2

# Digilent boards
from _nexys import nexys
from _atlys import atlys

from _xupv2p import xupv2p

from _ufo400 import ufo400
from _sx1 import sx1

# Papilio boards
from _papilio import pone


#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from myhdl import Signal,intbv
from ... import Clock,Reset
from ..._xilinx import XilinxFPGA


# Nexsys
nexys = XilinxFPGA('nexys')
nexys.add_port('clock', Clock(0, frequency=50e6), pins=('A8',))
nexys.add_port('reset', Reset(0,active=0,async=True), pins=(13,))

# misc user io
nexys.add_port('led', Signal(intbv(0)[8:]), 
               pins=('R16','P14','M13','N14',
                     'L12','M14','L13','L14',))
nexys.add_port('btn', Signal(intbv(0)[4:]),
               pins=('K12','K13','K14','J13',))
nexys.add_port('sw', Signal(intbv(0)[8:]),
               pins=('N16','M15','M16','L15',
                     'K15','K16','J16','N15',))

# FX2 USB interface
nexys.add_port('ifclk', Clock(0, frequency=48e6),
               pins=('T9',))
nexys.add_port('fdio', Signal(intbv(0)[8:]),
               pins=('R10','M10','P10','N10',
                     'P11','N11','P12','N12',))
nexys.add_port('addr', Signal(intbv(0)[2:]),
               pins=('M7','P5'))
nexys.add_port('flaga', Signal(bool(0)), pins=('N8',))
nexys.add_port('flagb', Signal(bool(0)), pins=('P7',))
nexys.add_port('flagc', Signal(bool(0)), pins=('N7',))
nexys.add_port('flagd', Signal(bool(0)), pins=('T8',))


# set the FPGA device on the board
nexys.set_device('spartan3e', 'XC3S500E', 'vq100', '-5')





from setuptools import setup,find_packages
from myhdl_tools import __version__

# The README is probably a little too long for the
# pipy stuff.
desc= \
"""
This Python package contains a mix-mash of tools to assist 
MyHDL designs.  Includes common components and scripts that 
are frequently used in complex digital designs.  The tools
assist in automating FPGA toolflows, simulations, and 
cosimulation.
"""

setup(name='myhdl_tools',
      version=__version__,
      author="Christopher Felton",
      author_email="chris.felton@gmail.com",
      license="LGPL",
      description="random collection of tools to support myhdl",
      keywords="myhdl FPGA tools",
      url="https://bitbucket.org/cfelton/myhdl_tools",
      packages = find_packages(),
      long_description=desc,
      test_suite = 'nose.collector',
      tests_require = 'nose',      
      )



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import inspect
from pprint import pprint
from random import randint
from copy import copy

import myhdl

from _mysigs import Clock

                
class _port(object):
    """ Generic pin/port definitions """
    
    def __init__(self, name, sig, pins, **kwargs):
        # add all the default attributes depeding on
        
        # fpga type
        self.name = name
        self.sig = sig
        self.pins = pins
        self.inuse = False

        # device specific arguments, when the pin assignment
        # list is created the following will be used to create
        # the UCF,TCL,etc input
        self.kwargs = kwargs
        
    def help(self):
        pass #useful help print!

class _fpga(object):

    def __init__(self):
        
        # add the default attributes (parameters)
        self.top = None
        self.tp = None
        self.top_name = None
        # keep track of all the thingies
        self.ports = {}
        self.clocks = {}

    def has_top(self):
        return self.top is not None
    
    def set_top(self, top, **params):
        self.top = top
        self.tp = params

    def pathexist(self, pth):
        """ create the working directory path if needed
        """
        if os.path.isfile(pth):
            pth,fn = os.path.split(pth)
        fpth = ''
        path_split = os.path.split(pth)
        for ppth in pth.split(os.path.sep):
            fpth = os.path.join(fpth,ppth)
            if not os.path.isdir(fpth):
                print("path create %s"%(fpth))
                os.mkdir(fpth)

        return os.path.isdir(pth)

    def _get_name(self, name=None):
        if name is None:
            if self.top_name is not None:
                name = self.top_name
            else:
                name = self.top.func_name

        return name

    def _scrub(self, filelist):
        """
        remove the generated files from the file list, if the
        toolflow is run subsequent times the converted files 
        are added each time (especially bad when running one
        conversion type and then the other)
        """
        print(filelist)
        name = self._get_name()
        fl = copy(filelist)
        for ii,ff in enumerate(fl):
            print(" ... %s"%(ff))
            if (ff == "%s.v"%(name) or 
                ff == "%s.vhd"%(name)):
                print("scrubbing %s"%(ff))
                filelist.remove(ff)
            elif "pck_myhdl" in ff:
                print("scrubbing %s"%(ff))
                filelist.remove(ff)

    def convert(self, name=None, to='verilog', **kwargs):
        """ convert the myhdl design to Verilog or VHDL and copy to working dir
        """
        assert to in ('verilog', 'vhdl'), \
               "Incorrect conversion type: %s" % (self.conversion_type)

        name = self._get_name(name)

        # get the top-level ports and parameters.
        pp = inspect.getargspec(self.top) # pp = portsparams

        # all of the arguments (no default values) are treated as
        # ports.  This doesn't mean it needs to be a port but it
        # is convention that ports are the arguments and parameters
        # are keyword arguments.  A parameter can exist in this
        # list but it can't be empty in conversion
        hdlports = {}
        if pp.defaults is not None:
            pi = len(pp.args)-len(pp.defaults)
        else:
            pi = len(pp.args)
        for pn in pp.args[:pi]:
            hdlports[pn] = None
        params = {}
        for ii,kw in enumerate(pp.args[pi:]):
            params[kw] = pp.defaults[ii]

        # see if any of the ports or parameters have been overridden
        if self.tp is not None:
            for k,v in self.tp.items():
                if hdlports.has_key(k):
                    hdlports[k] = v
            for k,v in self.tp.items():
                if params.has_key(k):
                    params[k] = v
                
        for k,v in kwargs.items():
            if hdlports.has_key(k):
                hdlports[k] = v
        for k,v in kwargs.items():
            if params.has_key(k):
                params[k] = v
        print('HDL PORTS:  ',hdlports)
        print('HDL PARAMS: ',params)
        
        # match the fpga ports to the hdl ports, not if a port is
        # a keyword argument in the top-level this will fail
        for port_name,port in self.ports.items():
            if hdlports.has_key(port_name):
                hdlports[port_name] = port.sig
                port.inuse = True

        for k,v in hdlports.items():
            assert v is not None, "Error unspecified port %s"%(k)
            
        # combine the ports and params
        pp = hdlports.copy()
        pp.update(params)
        
        # convert with the ports and parameters        
        if to.lower() == 'verilog':
            if name is not None:
                myhdl.toVerilog.name = name
            myhdl.toVerilog(self.top, **pp)
            self.name = name
            self.vfn = "%s.v"%(name)
        elif to.lower() == 'vhdl':
            if name is not None:
                myhdl.toVHDL.name = name
            myhdl.toVHDL(self.top, **pp)
            self.name = name
            self.vfn = "%s.vhd"%(name)
        else:
            raise ValueError("Incorrect conversion target %s"%(to))

        # make sure the working directory exists
        assert self.pathexist(self._path)

        # copy files etc to the working directory
        tbfn = 'tb_'+self.vfn
        ver = myhdl.__version__
        # remove special characters from the version
        for sp in ('.', '-', 'dev'):
            ver = ver.replace(sp,'')
        pckfn = 'pck_myhdl_%s.vhd'%(ver)
        for src in (self.vfn,tbfn,pckfn):
            dst = os.path.join(self._path,src)
            print('   checking %s'%(dst))
            if os.path.isfile(dst):
                print('   removing %s'%(dst))
                os.remove(dst)
            if os.path.isfile(src):
                print('   moving %s --> %s'%(src,self._path))
                shutil.move(src,self._path)

        if to.lower() == 'verilog':
            filelist = (self.vfn,)
        elif to.lower() == 'vhdl':
            filelist = (self.vfn, pckfn,)

        return filelist

    def add_port(self, name, sig, pins, **kwargs):
        if isinstance(pins, (str, int)):
            pins = (pins,)
        assert isinstance(pins, (list,tuple)), \
            "pins must be a list/tuple of pins (or a single str/int)"
        self.ports[name] =  _port(name, sig, pins, **kwargs)
        if isinstance(sig, Clock):
            self.clocks[name] = self.ports[name]

    def name_port(self, name, port, slice_=None, **kwargs):
        """Extract "slices" from a port
        Makes it possible to extract and name a single port (a range 
        of ports) out of a larger port array. This could be useful to 
        give meaningful names to single ports e.g. on Papilio Wings.

        Example: board.name_port('led', 'wingC', 7)
        where wingC is a 16-bit port array.

        To extract a port range the slice class has to be used. E.g.
        board.name_port('MSB', 'wingC', slice(16,8))"""

        p = self.ports[port]
        if slice_ is None:
            pins = p.pins
            sig = p.sig
        else:
            assert isinstance(slice_, (slice, int)), "slice_ has to be either None or a int/slice object"
            pins = p.pins[slice_]
            sig = myhdl.Signal(p.sig[slice_])
        kws = p.kwargs.copy()
        kws.update(kwargs)
        self.add_port(name, sig, pins, **kws)
            
    def run(self):
        pass



#
# Copyright (c) 2013 Christopher Felton
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from myhdl import Signal,intbv
from ... import Clock,Reset
from ..._xilinx import XilinxFPGA


# XUPV2P board (provided by Jos Huisken)
xupv2p = XilinxFPGA('xupv2p')
xupv2p.add_port('clock', Clock(0, frequency=100e6), pins=('AJ15',))
xupv2p.add_port('reset', Reset(0, active=0, async=True), pins=('AH5',))
xupv2p.add_port('led', Signal(intbv(0)[4:]),
               pins=('AC4','AC3','AA6','AA5'))
xupv2p.set_device('virtex2p','XC2VP30','ff896','-7')
